searchNodes=[{"type":"module","title":"Siblings","doc":"Bolerplate to effectively handle many long-lived entities of the same shape, driven by FSM. Siblings is a library to painlessly manage many uniform processes, all having the lifecycle and the FSM behind. Consider the service, that polls the market rates from several diffferent sources, allowing semi-automated trading based on predefined conditions. For each bid, the process is to be spawn, polling the external resources. Once the bid condition is met, the bid gets traded. With Siblings , one should implement Siblings.Worker.perform/3 callback, doing actual work and returning either :ok if no action should be taken, or {:transition, event, payload} to initiate the FSM transition. When the FSM get exhausted (reaches its end state,) both the performing process and the FSM itself do shut down. FSM instances leverage Finitomata library, which should be used alone if no recurrent perform should be accomplished or if the instances are not uniform. Typical code for the Siblings.Worker implementation would be as follows defmodule MyApp.Worker do @fsm &quot;&quot;&quot; born --&gt; |reject| rejected born --&gt; |bid| traded &quot;&quot;&quot; use Finitomata , @fsm def on_transition ( :born , :reject , _nil , payload ) do perform_rejection ( payload ) { :ok , :rejected , payload } end def on_transition ( :born , :bid , _nil , payload ) do perform_bidding ( payload ) { :ok , :traded , payload } end @behaviour Siblings.Worker @impl Siblings.Worker def perform ( state , id , payload ) def perform ( :born , id , payload ) do cond do time_to_bid? ( ) -&gt; { :transition , :bid , nil } stale? ( ) -&gt; { :transition , :reject , nil } true -&gt; :noop end end def perform ( :rejected , id , _payload ) do Logger . info ( &quot;The bid \#{ id } was rejected&quot; ) { :transition , :__end__ , nil } end def perform ( :traded , id , _payload ) do Logger . info ( &quot;The bid \#{ id } was traded&quot; ) { :transition , :__end__ , nil } end end Now it can be used as shown below { :ok , pid } = Siblings . start_link ( ) Siblings . start_child ( MyApp.Worker , &quot;Bid1&quot; , %{ } , interval : 1_000 ) Siblings . start_child ( MyApp.Worker , &quot;Bid2&quot; , %{ } , interval : 1_000 ) ... The above would spawn two processes, checking the conditions once per a second ( interval ,) and manipulating the underlying FSM to walk through the bids’ lifecycles. Worker’s interval might be reset with GenServer.cast(pid, {:reset, interval}) and the message might be casted to it with GenServer.call(pid, {:message, message}) . For the latter to work, the optional callback on_call/2 must be implemented. Sidenote: Normally, Siblings supervisor would be put into the supervision tree of the target application.","ref":"Siblings.html"},{"type":"function","title":"Siblings.call/3","doc":"Performs a GenServer.call/3 on the named worker.","ref":"Siblings.html#call/3"},{"type":"function","title":"Siblings.child_spec/1","doc":"Returns the child spec for the named or unnamed Siblings process. Useful when many Siblings processes are running simultaneously.","ref":"Siblings.html#child_spec/1"},{"type":"function","title":"Siblings.multi_call/2","doc":"Performs a GenServer.call/3 on all the workers.","ref":"Siblings.html#multi_call/2"},{"type":"function","title":"Siblings.multi_transition/3","doc":"Initiates the transition of all the workers.","ref":"Siblings.html#multi_transition/3"},{"type":"function","title":"Siblings.payload/2","doc":"Returns the payload of FSM behind the named worker.","ref":"Siblings.html#payload/2"},{"type":"function","title":"Siblings.reset/3","doc":"Resets the the named worker’s interval.","ref":"Siblings.html#reset/3"},{"type":"function","title":"Siblings.start_child/4","doc":"Starts the supervised child under the PartitionSupervisor .","ref":"Siblings.html#start_child/4"},{"type":"function","title":"Siblings.start_link/1","doc":"Starts the supervision subtree, holding the PartitionSupervisor . This is the main entry point of Siblings . It starts the supervision tree, holding the partitioned DynamicSupervisor s, the lookup to access children, and the optional set of workers to start immediately. Siblings are fully controlled by FSM instances. Children are added using Siblings.Lookup interface methods which go all the way through underlying FSM implementation. opts might include: name: atom() which is a name of the Siblings instance, defaults to Siblings workers: list() the list of the workers to start imminently upon Siblings start throttler: keyword() the throttler options, see Siblings.Throttler for details die_with_children: true | false | (-&gt; :ok) | {(-&gt; :ok), timeout} shutdown the process when there is no more active child, defaults to false (if a function of arity 0 is given, it’ll be called before the process shuts down) callbacks: list() the list of the handler to call back upon Lookup transitions","ref":"Siblings.html#start_link/1"},{"type":"function","title":"Siblings.state/3","doc":"Returns the state of the Siblings instance itself, of the named worker, or the named worker’s underlying FSM , depending on the first argument.","ref":"Siblings.html#state/3"},{"type":"function","title":"Siblings.states/1","doc":"Returns the states of all the workers as a map.","ref":"Siblings.html#states/1"},{"type":"function","title":"Siblings.transition/4","doc":"Initiates the transition of the named worker.","ref":"Siblings.html#transition/4"},{"type":"type","title":"Siblings.start_options/0","doc":"","ref":"Siblings.html#t:start_options/0"},{"type":"type","title":"Siblings.worker/0","doc":"","ref":"Siblings.html#t:worker/0"},{"type":"module","title":"Siblings.Lookup","doc":"The instance of FSM backed up by Finitomata . FSM representation graph TD idle --&gt; |initialize| ready ready --&gt; |start_child| ready ready --&gt; |delete_child| ready ready --&gt; |terminate| terminated","ref":"Siblings.Lookup.html"},{"type":"function","title":"Siblings.Lookup.__config__/1","doc":"Getter for the internal compiled in FSM information.","ref":"Siblings.Lookup.html#__config__/1"},{"type":"function","title":"Siblings.Lookup.all/1","doc":"Returns all the workers running under this Siblings instance as a map %{Siblings.Worker.id() =&gt; pid()} . This map might be really huge when there are many processes managed.","ref":"Siblings.Lookup.html#all/1"},{"type":"function","title":"Siblings.Lookup.child_spec/1","doc":"Returns a specification to start this module under a supervisor. See Supervisor .","ref":"Siblings.Lookup.html#child_spec/1"},{"type":"macro","title":"Siblings.Lookup.config/1","doc":"The convenient macro to allow using states in guards, returns a compile-time list of states for Siblings.Lookup .","ref":"Siblings.Lookup.html#config/1"},{"type":"function","title":"Siblings.Lookup.del/2","doc":"Removes the reference for the naturally terminated child from the workers map through :delete_child transition with all the respective callbacks.","ref":"Siblings.Lookup.html#del/2"},{"type":"function","title":"Siblings.Lookup.get/3","doc":"Returns the pid of the single dynamically supervised worker by its id .","ref":"Siblings.Lookup.html#get/3"},{"type":"function","title":"Siblings.Lookup.put/2","doc":"Initiates the :start_child transition with all the respective callbacks to add a new child to the supervised list.","ref":"Siblings.Lookup.html#put/2"},{"type":"function","title":"Siblings.Lookup.start_link/1","doc":"Starts an FSM alone with name and payload given. Usually one does not want to call this directly, the most common way would be to start a Finitomata supervision tree or even better embed it into the existing supervision tree and start FSM with Finitomata.start_fsm/3 passing Siblings.Lookup as the first parameter.","ref":"Siblings.Lookup.html#start_link/1"},{"type":"type","title":"Siblings.Lookup.state/0","doc":"Kind of event which might be send to initiate the transition. FSM representation graph TD idle --&gt; |initialize| ready ready --&gt; |start_child| ready ready --&gt; |delete_child| ready ready --&gt; |terminate| terminated","ref":"Siblings.Lookup.html#t:state/0"},{"type":"module","title":"Siblings.Test.Callable","doc":"The instance of FSM backed up by Finitomata . FSM representation graph TD s1 --&gt; |to_s2| s2 s2 --&gt; |to_s3| s3","ref":"Siblings.Test.Callable.html"},{"type":"function","title":"Siblings.Test.Callable.__config__/1","doc":"Getter for the internal compiled in FSM information.","ref":"Siblings.Test.Callable.html#__config__/1"},{"type":"function","title":"Siblings.Test.Callable.child_spec/1","doc":"Returns a specification to start this module under a supervisor. See Supervisor .","ref":"Siblings.Test.Callable.html#child_spec/1"},{"type":"macro","title":"Siblings.Test.Callable.config/1","doc":"The convenient macro to allow using states in guards, returns a compile-time list of states for Siblings.Test.Callable .","ref":"Siblings.Test.Callable.html#config/1"},{"type":"function","title":"Siblings.Test.Callable.start_link/1","doc":"Starts an FSM alone with name and payload given. Usually one does not want to call this directly, the most common way would be to start a Finitomata supervision tree or even better embed it into the existing supervision tree and start FSM with Finitomata.start_fsm/3 passing Siblings.Test.Callable as the first parameter.","ref":"Siblings.Test.Callable.html#start_link/1"},{"type":"type","title":"Siblings.Test.Callable.state/0","doc":"Kind of event which might be send to initiate the transition. FSM representation graph TD s1 --&gt; |to_s2| s2 s2 --&gt; |to_s3| s3","ref":"Siblings.Test.Callable.html#t:state/0"},{"type":"module","title":"Siblings.Test.FSM","doc":"The instance of FSM backed up by Finitomata . FSM representation graph TD s1 --&gt; |to_s2| s2 s1 --&gt; |to_s3| s3","ref":"Siblings.Test.FSM.html"},{"type":"function","title":"Siblings.Test.FSM.__config__/1","doc":"Getter for the internal compiled in FSM information.","ref":"Siblings.Test.FSM.html#__config__/1"},{"type":"function","title":"Siblings.Test.FSM.child_spec/1","doc":"Returns a specification to start this module under a supervisor. See Supervisor .","ref":"Siblings.Test.FSM.html#child_spec/1"},{"type":"macro","title":"Siblings.Test.FSM.config/1","doc":"The convenient macro to allow using states in guards, returns a compile-time list of states for Siblings.Test.FSM .","ref":"Siblings.Test.FSM.html#config/1"},{"type":"function","title":"Siblings.Test.FSM.start_link/1","doc":"Starts an FSM alone with name and payload given. Usually one does not want to call this directly, the most common way would be to start a Finitomata supervision tree or even better embed it into the existing supervision tree and start FSM with Finitomata.start_fsm/3 passing Siblings.Test.FSM as the first parameter.","ref":"Siblings.Test.FSM.html#start_link/1"},{"type":"type","title":"Siblings.Test.FSM.state/0","doc":"Kind of event which might be send to initiate the transition. FSM representation graph TD s1 --&gt; |to_s2| s2 s1 --&gt; |to_s3| s3","ref":"Siblings.Test.FSM.html#t:state/0"},{"type":"module","title":"Siblings.Test.NoPerform","doc":"The instance of FSM backed up by Finitomata . FSM representation graph TD s1 --&gt; |to_s2| s2 s2 --&gt; |in_s2| s2 s2 --&gt; |to_s3| s3","ref":"Siblings.Test.NoPerform.html"},{"type":"function","title":"Siblings.Test.NoPerform.__config__/1","doc":"Getter for the internal compiled in FSM information.","ref":"Siblings.Test.NoPerform.html#__config__/1"},{"type":"function","title":"Siblings.Test.NoPerform.child_spec/1","doc":"Returns a specification to start this module under a supervisor. See Supervisor .","ref":"Siblings.Test.NoPerform.html#child_spec/1"},{"type":"macro","title":"Siblings.Test.NoPerform.config/1","doc":"The convenient macro to allow using states in guards, returns a compile-time list of states for Siblings.Test.NoPerform .","ref":"Siblings.Test.NoPerform.html#config/1"},{"type":"function","title":"Siblings.Test.NoPerform.start_link/1","doc":"Starts an FSM alone with name and payload given. Usually one does not want to call this directly, the most common way would be to start a Finitomata supervision tree or even better embed it into the existing supervision tree and start FSM with Finitomata.start_fsm/3 passing Siblings.Test.NoPerform as the first parameter.","ref":"Siblings.Test.NoPerform.html#start_link/1"},{"type":"type","title":"Siblings.Test.NoPerform.state/0","doc":"Kind of event which might be send to initiate the transition. FSM representation graph TD s1 --&gt; |to_s2| s2 s2 --&gt; |in_s2| s2 s2 --&gt; |to_s3| s3","ref":"Siblings.Test.NoPerform.html#t:state/0"},{"type":"module","title":"Siblings.Test.WorkerFSM","doc":"The instance of FSM backed up by Finitomata . FSM representation graph TD s1 --&gt; |to_s2| s2 s2 --&gt; |in_s2| s2 s2 --&gt; |to_s3| s3","ref":"Siblings.Test.WorkerFSM.html"},{"type":"function","title":"Siblings.Test.WorkerFSM.__config__/1","doc":"Getter for the internal compiled in FSM information.","ref":"Siblings.Test.WorkerFSM.html#__config__/1"},{"type":"function","title":"Siblings.Test.WorkerFSM.child_spec/1","doc":"Returns a specification to start this module under a supervisor. See Supervisor .","ref":"Siblings.Test.WorkerFSM.html#child_spec/1"},{"type":"macro","title":"Siblings.Test.WorkerFSM.config/1","doc":"The convenient macro to allow using states in guards, returns a compile-time list of states for Siblings.Test.WorkerFSM .","ref":"Siblings.Test.WorkerFSM.html#config/1"},{"type":"function","title":"Siblings.Test.WorkerFSM.start_link/1","doc":"Starts an FSM alone with name and payload given. Usually one does not want to call this directly, the most common way would be to start a Finitomata supervision tree or even better embed it into the existing supervision tree and start FSM with Finitomata.start_fsm/3 passing Siblings.Test.WorkerFSM as the first parameter.","ref":"Siblings.Test.WorkerFSM.html#start_link/1"},{"type":"type","title":"Siblings.Test.WorkerFSM.state/0","doc":"Kind of event which might be send to initiate the transition. FSM representation graph TD s1 --&gt; |to_s2| s2 s2 --&gt; |in_s2| s2 s2 --&gt; |to_s3| s3","ref":"Siblings.Test.WorkerFSM.html#t:state/0"},{"type":"module","title":"Siblings.Throttler","doc":"The internal definition of the call to throttle. Siblings.Throttler.call/3 is a blocking call similar to GenServer.call/3 , but served by the underlying GenStage producer-consumer pair. Despite this implementation of throttling based on GenStage is provided mostly for internal needs, it is generic enough to use wherever. Use the childspec {Siblings.Throttler, name: name, initial: [], max_demand: 3, interval: 1_000} to start a throttling process and Siblings.Throttler.call/3 to perform throttled synchronous calls from different processes.","ref":"Siblings.Throttler.html"},{"type":"function","title":"Siblings.Throttler.call/3","doc":"Synchronously executes the function, using throttling based on GenStage . This function has a default timeout :infinity because of its nature (throttling is supposed to take a while,) but it might be passed as the third argument in a call to call/3 . If a list of functions is given, executes all of them in parallel, collects the results, and then returns them to the caller. The function might be given as t:Siblings.Throttler.t() or in a simplified form as {function_of_arity_1, arg} .","ref":"Siblings.Throttler.html#call/3"},{"type":"function","title":"Siblings.Throttler.child_spec/1","doc":"Returns a specification to start this module under a supervisor. See Supervisor .","ref":"Siblings.Throttler.html#child_spec/1"},{"type":"function","title":"Siblings.Throttler.start_link/1","doc":"Starts the throttler with the underlying producer-consumer stages. Accepted options are: name the base name for the throttler to be used in calls to call/3 initial the initial load of requests (avoid using it unless really needed) max_demand , initial the options to be passed directly to GenStage ’s consumer","ref":"Siblings.Throttler.html#start_link/1"},{"type":"type","title":"Siblings.Throttler.t/0","doc":"The in/out parameter for calls to Siblings.Throttler.call/3","ref":"Siblings.Throttler.html#t:t/0"},{"type":"type","title":"Siblings.Throttler.throttlee/0","doc":"The simplified in parameter for calls to Siblings.Throttler.call/3","ref":"Siblings.Throttler.html#t:throttlee/0"},{"type":"behaviour","title":"Siblings.Worker","doc":"The worker for the single sibling process.","ref":"Siblings.Worker.html"},{"type":"callback","title":"Siblings.Worker.finitomata/0","doc":"The Finitomata FSM implementation module. It will be used internally to carry the state of FSM. If not implemented, this module itself will be considered an FSM implementation.","ref":"Siblings.Worker.html#c:finitomata/0"},{"type":"callback","title":"Siblings.Worker.on_call/2","doc":"The handler for the routed message from Siblings.InternalWorker.handle_call({:message, any()}) .","ref":"Siblings.Worker.html#c:on_call/2"},{"type":"callback","title":"Siblings.Worker.on_init/1","doc":"The function to re-initialize FSM after crash.","ref":"Siblings.Worker.html#c:on_init/1"},{"type":"callback","title":"Siblings.Worker.perform/3","doc":"The callback to be implemented in each and every worker. It will be called internally continiously by the internal worker process, transitioning the underlying FSM according to the return value.","ref":"Siblings.Worker.html#c:perform/3"},{"type":"type","title":"Siblings.Worker.call_result/0","doc":"Value, returned from on_call/2 callback","ref":"Siblings.Worker.html#t:call_result/0"},{"type":"type","title":"Siblings.Worker.id/0","doc":"Identifier of the worker process","ref":"Siblings.Worker.html#t:id/0"},{"type":"type","title":"Siblings.Worker.message/0","doc":"Message to be sent to the worker process","ref":"Siblings.Worker.html#t:message/0"},{"type":"type","title":"Siblings.Worker.payload/0","doc":"Payload associated with the worker","ref":"Siblings.Worker.html#t:payload/0"},{"type":"module","title":"Siblings.InternalWorker","doc":"The internal process to manage Siblings.Worker subsequent runs along with its FSM .","ref":"Siblings.InternalWorker.html"},{"type":"function","title":"Siblings.InternalWorker.child_spec/1","doc":"Returns a specification to start this module under a supervisor. See Supervisor .","ref":"Siblings.InternalWorker.html#child_spec/1"},{"type":"type","title":"Siblings.InternalWorker.options/0","doc":"Allowed options in a call to start_link/4","ref":"Siblings.InternalWorker.html#t:options/0"},{"type":"module","title":"Siblings.InternalWorker.State","doc":"The state of the worker.","ref":"Siblings.InternalWorker.State.html"},{"type":"type","title":"Siblings.InternalWorker.State.t/0","doc":"","ref":"Siblings.InternalWorker.State.html#t:t/0"},{"type":"extras","title":"Siblings","doc":"The partitioned dynamic supervision of FSM-backed workers.","ref":"readme.html"},{"type":"extras","title":"Siblings - Usage","doc":"Siblings is a library to painlessly manage many uniform processes, all having the lifecycle and the FSM behind. Consider the service, that polls the market rates from several diffferent sources, allowing semi-automated trading based on predefined conditions. For each bid, the process is to be spawn, polling the external resources. Once the bid condition is met, the bid gets traded. With Siblings , one should implement Siblings.Worker.perform/3 callback, doing actual work and returning either :ok if no action should be taken, or {:transition, event, payload} to initiate the FSM transition. When the FSM get exhausted (reaches its end state,) both the performing process and the FSM itself do shut down. FSM instances leverage Finitomata library, which should be used alone if no recurrent perform should be accomplished or if the instances are not uniform. Typical code for the Siblings.Worker implementation would be as follows defmodule MyApp.Worker do @fsm &quot;&quot;&quot; born --&gt; |reject| rejected born --&gt; |bid| traded &quot;&quot;&quot; use Finitomata , @fsm def on_transition ( :born , :reject , _nil , payload ) do perform_rejection ( payload ) { :ok , :rejected , payload } end def on_transition ( :born , :bid , _nil , payload ) do perform_bidding ( payload ) { :ok , :traded , payload } end @behaviour Siblings.Worker @impl Siblings.Worker def perform ( state , id , payload ) def perform ( :born , id , payload ) do cond do time_to_bid? ( ) -&gt; { :transition , :bid , nil } stale? ( ) -&gt; { :transition , :reject , nil } true -&gt; :noop end end def perform ( :rejected , id , _payload ) do Logger . info ( &quot;The bid \#{ id } was rejected&quot; ) { :transition , :__end__ , nil } end def perform ( :traded , id , _payload ) do Logger . info ( &quot;The bid \#{ id } was traded&quot; ) { :transition , :__end__ , nil } end end Now it can be used as shown below { :ok , pid } = Siblings . start_link ( ) Siblings . start_child ( MyApp.Worker , &quot;Bid1&quot; , %{ } , interval : 1_000 ) Siblings . start_child ( MyApp.Worker , &quot;Bid2&quot; , %{ } , interval : 1_000 ) ... The above would spawn two processes, checking the conditions once per a second ( interval ,) and manipulating the underlying FSM to walk through the bids’ lifecycles. Worker’s interval might be reset with GenServer.cast(pid, {:reset, interval}) and the message might be casted to it with GenServer.call(pid, {:message, message}) . For the latter to work, the optional callback on_call/2 must be implemented. Sidenote: Normally, Siblings supervisor would be put into the supervision tree of the target application.","ref":"readme.html#usage"},{"type":"extras","title":"Siblings - Installation","doc":"def deps do [ { :siblings , &quot;~&gt; 0.1&quot; } ] end","ref":"readme.html#installation"},{"type":"extras","title":"Siblings - Changelog","doc":"0.11.3 — OTP26 ready 0.11.2 — [FIX] wrong specs for start_link/1 and child_spec/1 0.11.1 — upgraded to Finitomata ( v0.11.0 ) 0.11.0 — throttler → generic + on perform 0.10.3 — accept {(any() -&gt; :ok), timeout} as die_with_children , write-only InternalState 0.10.2 — accept (any() -&gt; :ok) as die_with_children option as a callback 0.10.0 — die_with_children: boolean() option 0.8.2 — updated with last finitomata compiler 0.7.0 — Siblings.state/{0,1,2,3} + update to Finitoma 0.7 0.5.1 — allow {:reschedule, non_neg_integer()} return from perform/3 0.5.0 — use FSM for the Sibling.Lookup 0.4.3 — accept hibernate?: boolean parameter in call to Siblings.start_child/4 to hibernate children 0.4.2 — accept workers: in call to Siblings.child_spec/1 to statically initialize Siblings 0.4.1 — [BUG] many named Siblings instances 0.4.0 — Siblings.{multi_call/2, multi_transition/3} 0.3.3 — Siblings.{state/1, payload/2} 0.3.2 — Siblings.{call/3, reset/3, transition/4} 0.3.1 — retrieve childrens as both map and list 0.3.0 — GenServer.cast(pid, {:reset, interval}) and GenServer.call(pid, {:message, message}) 0.2.0  — Fast Worker lookup 0.1.0  — Initial MVP","ref":"readme.html#changelog"},{"type":"extras","title":"Siblings - Documentation","doc":"","ref":"readme.html#documentation"}]